import imagehash
import io
import numpy as np
import pymongo
import requests
from PIL import Image

from salesvision.server_estimation import EstimationTemplate
from salesvision.metrics import set2set_similarity_metric as similarity_metric
from salesvision.image_utils import read_image, compute_image_hash


class LocalDatabaseWithIntegratedEstimator(object):
    """Class for local database usage with paths"""
    def __init__(self,
                 estimator: EstimationTemplate,
                 database_name: str = 'ClothesBase',
                 with_hash_indexing: bool = True,
                 metric: callable = similarity_metric):
        """
        Class constructor
        Args:
            estimator: Estimation engine
            database_name: name of created local database, if database with
                same name has already been created,
                then it is used for solution
            with_hash_indexing: preventing occurrence duplicates by image hash
        """
        self.client = pymongo.MongoClient()
        self.db_name = database_name
        self.db = self.client[database_name]
        self.collection = self.db.images
        self.with_hash_indexing = with_hash_indexing
        self.estimator = estimator
        self.metric = metric

    def _order_database(self, prediction: dict) -> list:
        """
        Find similarity elements in base by prediction results
        Args:
            prediction: system prediction

        Returns:
            List of indexes ordered by metric similarity
        """
        indexes_with_metric = []

        for elem in self.collection.find():
            indexes_with_metric.append(
                (
                    elem['_id'],
                    self.metric(prediction, elem)
                )
            )

        indexes_with_metric.sort(key=lambda x: x[1], reverse=True)

        return indexes_with_metric

    def _in_database(self, image: Image, image_hash: str) -> tuple:
        """

        Args:
            image:
            image_hash:

        Returns:

        """
        if self.collection.count_documents({'image_hash': image_hash}) > 0:
            for elem in self.collection.find(
                    {'image_hash': image_hash}):
                sample_img = read_image(elem['local_path'])
                sample_img = np.array(sample_img)
                img = np.array(image)

                if np.abs(sample_img - img).sum() == 0:
                    return True, elem

        return False, None

    def insert(self, image_object) -> pymongo.collection.ObjectId:
        """
        Insert image description to database
        Args:
            image_object: image path on local machine or URL link or Array

        Returns:
            Inserted element ID in database
        """
        image = read_image(image_object)
        img_hash = compute_image_hash(image)

        if self.with_hash_indexing:
            if self._in_database(image, img_hash)[0]:
                raise RuntimeWarning('Database has contained image already')

        results = self.estimator(image_object)

        insert_data = {
            'local_path': image_object if type(image_object) is str else '',
            'image_hash': img_hash,
            'prediction': results
        }

        return self.collection.insert_one(insert_data).inserted_id

    def search(self,
               image_object,
               top_k: int = 5,
               return_scores: bool = False) -> list:
        """
        Search top k similar samples indexes from local database
        Args:
            image_object: image path on local machine or URL link or Array
            top_k: count of found samples
            return_scores: return list of tuples with indexes and metric scores

        Returns:
            Indexes of top k similar samples from database
        """
        image = read_image(image_object)
        img_hash = compute_image_hash(image)

        need_predict = True
        prediction = None
        if self.with_hash_indexing:
            checking_result = self._in_database(image, img_hash)
            if checking_result[0]:
                prediction = checking_result[1]
                need_predict = False
                print('From base')

        if need_predict:
            prediction = {'prediction': self.estimator(image_object)}

        if prediction['prediction']['status'] != '0':
            raise RuntimeError(
                'System couldn\'t process image, error code: {}'.format(
                    prediction['prediction']['status']
                )
            )

        search_result = self._order_database(prediction)
        if not return_scores:
            search_result = [elem[0] for elem in search_result]

        if len(search_result) < top_k:
            return search_result
        return search_result[:top_k]

    def drop_collection(self):
        """
        Erase database
        Returns:
            None
        """
        return self.collection.drop()

    def get_images_by_indexes(self, indexes: list) -> list:
        """
        Return Pillow images list from base by indexes list
        Args:
            indexes: indexes list

        Returns:
            Pillow images list

        """
        return [
            read_image(
                self.collection.find_one(
                    {'_id': idx[0] if type(idx) is tuple else idx}
                )['local_path']
            )
            for idx in indexes
        ]


class SimpleLocalDatabase(object):
    """Class for local database usage"""
    def __init__(self,
                 database_name: str = 'ClothesBase',
                 metric: callable = similarity_metric):
        """
        Class constructor
        Args:
            database_name: name of created local database, if database with
                same name has already been created,
                then it is used for solution
        """
        self.client = pymongo.MongoClient()
        self.db_name = database_name
        self.db = self.client[database_name]
        self.collection = self.db.images
        self.metric = metric

    def _order_database(self, prediction: dict) -> list:
        """
        Find similarity elements in base by prediction results
        Args:
            prediction: system prediction

        Returns:
            List of indexes ordered by metric similarity
        """
        indexes_with_metric = []

        for elem in self.collection.find():
            indexes_with_metric.append(
                (
                    elem['_id'],
                    self.metric(prediction, elem)
                )
            )

        indexes_with_metric.sort(key=lambda x: x[1], reverse=True)

        return indexes_with_metric

    def insert(self, prediction: dict) -> pymongo.collection.ObjectId:
        """
        Insert image description to database
        Args:
            prediction: system prediction

        Returns:
            Inserted element ID in database
        """
        insert_data = {
            'local_path': '',
            'image_hash': '',
            'prediction': prediction
        }

        return self.collection.insert_one(insert_data).inserted_id

    def search(self,
               prediction: dict,
               top_k: int = 5,
               return_scores: bool = False) -> list:
        """
        Search top k similar samples indexes from local database
        Args:
            prediction: system prediction
            top_k: count of found samples
            return_scores: return list of tuples with indexes and metric scores

        Returns:
            Indexes of top k similar samples from database
        """
        if prediction['status'] != '0':
            raise RuntimeError(
                'System couldn\'t process image, error code: {}'.format(
                    prediction['prediction']['status']
                )
            )

        base_prediction = {'prediction': prediction}

        search_result = self._order_database(base_prediction)
        if not return_scores:
            search_result = [elem[0] for elem in search_result]

        if len(search_result) < top_k:
            return search_result
        return search_result[:top_k]

    def drop_collection(self):
        """
        Erase database
        Returns:
            None
        """
        return self.collection.drop()



import imagehash
import io
import numpy as np
from PIL import Image
import requests


def read_image(image_object):
    """
    Read image by path or URL or Array
    Args:
        image_object: image path on local machine or URL link or Array

    Returns:
        Image in Pillow format or None value if invalid input image object
    """
    if type(image_object) is np.ndarray:
        return Image.fromarray(image_object)

    if type(image_object) is str and len(image_object) == 0:
        return None

    if len(image_object) > 7 and \
            ('http://' == image_object[:7] or 'https://' == image_object[:8]):
        response = requests.get(image_object)
        image_bin = io.BytesIO(response.content)
    else:
        image_bin = open(image_object, "rb")

    image = Image.open(image_bin)

    return image


def compute_image_hash(image: Image):
    """
    Compute image hash
    Args:
        image: image in Pillow format

    Returns:
        string with image hash
    """
    return str(imagehash.average_hash(image))



import numpy as np


def set2set_similarity_metric(pred1: dict, pred2: dict) -> float:
    """
    Evaluate similarity between two predictions,
    than greater value, than more similar things clothes
    Args:
        pred1: system prediction
        pred2: system prediction

    Returns:
        Similarity score
    """
    if pred1['prediction']['status'] != '0' or \
            pred2['prediction']['status'] != '0':
        return -1E+5

    score = 0
    alpha = [1.0, 0.1]
    beta = [[1.0, 0.8], [0.5, 1.0]]
    matches_count = 0
    matched_items = []

    for item1 in pred1['prediction']['result']:
        for item2 in pred2['prediction']['result']:
            if item2 in matched_items:
                continue

            matched_tags_count = 0
            if 'tags' in item1.keys() and 'tags' in item2.keys():
                for category in item1['tags'].keys():
                    matched_tags_count += int(
                        item1['tags'][category] == item2['tags'][category]
                    )
            matched_tags_rate = matched_tags_count / 7.0

            colors_similarity_score = 2.0 - np.linalg.norm(
                np.array(item1['color_embedding']) -
                np.array(item2['color_embedding'])
            )

            if item1['category'] == item2['category']:
                items_score = alpha[0]*(
                        beta[0][0]*matched_tags_rate +
                        beta[0][1]*colors_similarity_score
                )
                matched_items.append(item2)
            else:
                items_score = alpha[1] * (
                        beta[1][0] * matched_tags_rate +
                        beta[1][1] * colors_similarity_score
                )

            matches_count += 1
            score += items_score

    score = score / matches_count
    return score


def set2element_similarity_metric(
        pred1: dict, pred2: dict,
        alpha: float = 0.5, beta: float = 0.9) -> float:
    """
    Evaluate similarity between two predictions,
    than greater value, than more similar things clothes
    Args:
        pred1: system prediction
        pred2: system prediction
        alpha:
        beta:

    Returns:
        Similarity score
    """
    if pred1['prediction']['status'] != '0' or \
            pred2['prediction']['status'] != '0':
        return -1E+5

    assert 0 <= alpha <= 1
    assert 0 <= beta <= 1

    matches_count = 0
    match_scores = []

    for item1 in pred1['prediction']['result']:
        for item2 in pred2['prediction']['result']:
            matched_tags_count = 0
            if 'tags' in item1.keys() and 'tags' in item2.keys():
                for category in item1['tags'].keys():
                    matched_tags_count += int(
                        item1['tags'][category] == item2['tags'][category]
                    )
            matched_tags_rate = matched_tags_count / 7.0

            colors_similarity_score = (2.0 - np.linalg.norm(
                np.array(item1['color_embedding']) -
                np.array(item2['color_embedding'])
            )) / 2.0

            item_score = \
                matched_tags_rate * alpha \
                + colors_similarity_score * (1.0 - alpha)

            matches_count += 1
            match_scores.append(
                item_score * beta
                if item1['category'] == item2['category'] else
                item_score * (1.0 - beta)
            )

    match_scores.sort(reverse=True)

    score = match_scores[0]
    return score



import io
import requests
from PIL import Image

from salesvision.image_utils import read_image


class EstimationTemplate(object):
    def __call__(self, image_object) -> dict:
        raise NotImplementedError()


class ServerEstimator(EstimationTemplate):
    def __init__(self,
                 server_ip: str = '87.117.25.190',
                 server_port: int = 5015):
        """
        Class constructor
        Args:
            server_ip: estimation server id address in format *.*.*.*
            server_port: estimation server port
        """
        self.address = 'http://{}:{}'.format(server_ip, server_port)

        response = requests.get(
            url='{}/api/test/ping/'.format(self.address)
        )

        if response.status_code != 200:
            raise RuntimeError(
                'Can\'t connect to server: {}'.format(self.address)
            )

    def _analyze_image(self, image: Image):
        """
        Server inference on image
        Args:
            image: image in Pillow format

        Returns:
            Dict with prediction
        """
        img_byte_array = io.BytesIO()
        image.save(img_byte_array, format='JPEG')
        img_byte_array = img_byte_array.getvalue()

        response = requests.post(
            url='{}/api/inference/client_fashion_analysis/'.format(
                self.address
            ),
            files={"file": ("filename", img_byte_array, "image/jpeg")}
        )

        if response.status_code != 200:
            raise RuntimeError(
                'Request fail with code: {}'.format(response.status_code)
            )

        results = response.json()

        return results

    def __call__(self, image_object) -> dict:
        """
        Insert image description to database
        Args:
            image_object: image path on local machine or URL link or Array

        Returns:
            Prediction in dictionary format
        """
        image = read_image(image_object)

        return self._analyze_image(image)
